/*
 * Copyright (c) 1994, 1996 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <TopoFace/fgeomobjs.h>
#include <Dem/usgsdem.h>
#include <OS/math.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#if defined(sun)
extern "C" int pclose(FILE*);
#endif

UsgsDem::UsgsDem() 
: DEM()
{
    memset(name, 0, (size_t)41);
    memset(comment, 0, (size_t)41);
    memset(processcode, 0, (size_t)2);
    memset(section, 0, (size_t)4);
    memset(mcorigin, 0, (size_t)5);
    memset(demlevel, 0, (size_t)7);
    memset(elevpattern, 0, (size_t)7);
    memset(referencesys, 0, (size_t)7);
    memset(refzone, 0, (size_t)7);
    memset(projparams, 0, (size_t)361);
    memset(coordunits, 0, (size_t)7);
    memset(elevunits, 0, (size_t)7);
    memset(nsides, 0, (size_t)7);
    memset(cornercoords, 0, (size_t)193);
    memset(elevminmax, 0, (size_t)49);
    memset(angle, 0, (size_t)25);
    memset(acccode, 0, (size_t)7);
    memset(resolutionunits, 0, (size_t)37);
    memset(numrowscolumns, 0, (size_t)13);
    memset(largestcontour, 0, (size_t)6);
    memset(largestunits, 0, (size_t)2);
    memset(smallestcontour, 0, (size_t)6);
    memset(smallestunits, 0, (size_t)2);
    memset(sourcedate, 0, (size_t)5);
    memset(revdate, 0, (size_t)5);
    memset(insprevflag, 0, (size_t)2);
    memset(validationflag, 0, (size_t)2);
    memset(suspvoidflags, 0, (size_t)3);
    memset(verticaldatum, 0, (size_t)3);
    memset(horizontaldatum, 0, (size_t)3);
    memset(dataedition, 0, (size_t)5);
    memset(percentvoid, 0, (size_t)5);
    level = NULL_QLEVEL;
    refsys = GEOGRAPHIC;
    horz_units = RADIANS;
    vert_units = RADIANS;
    vdatum = NULL_VDATUM;
    hdatum = NULL_HDATUM;
    startrow = NULL;
    endrow = NULL;
}

UsgsDem::~UsgsDem() {
    delete [] startrow;
    delete [] endrow;
}

void UsgsDem::read(char* path) {
    FILE* f;
    int p = 0;
    int len = strlen(path);
    if (path[len-1]=='Z' && path[len-2]=='.') {
	char* cmd = new char[len+6];
	strcpy(cmd, "zcat ");
	strcat(cmd, path);
	if ((f = popen(cmd, "r")) == NULL) {
	    printf("popen failed\n");
	    return;
	}
	delete cmd;
	p = 1;
    }
    else if (path[len-1]=='z' && path[len-2]=='g' && path[len-3]=='.') {
	char* cmd = new char[len+11];
	strcpy(cmd, "gunzip -c ");
	strcat(cmd, path);
	if ((f = popen(cmd, "r")) == NULL) {
	    printf("popen failed\n");
	    return;
	}
	delete cmd;
	p = 1;
    }
    else if ((f = fopen(path, "r")) == NULL) {
	printf("fopen failed\n");
	return;
    }

    char buf[1025];

    if (fgets(buf, 1025, f) == NULL) {
	printf("error in fgets\n");
	return;
    }
    memcpy(name, &buf[0], (size_t)40);
    memcpy(comment, &buf[40], (size_t)40);
    memcpy(processcode, &buf[135], (size_t)1);
    memcpy(section, &buf[137], (size_t)3);
    memcpy(mcorigin, &buf[140], (size_t)4);
    memcpy(demlevel, &buf[144], (size_t)6);
    sscanf(demlevel, "%d", &level);
    memcpy(elevpattern, &buf[150], (size_t)6);
    memcpy(referencesys, &buf[156], (size_t)6);
    sscanf(referencesys, "%d", &refsys);
    if (refsys == STATE_PLANE)
	printf("warning: can't handle state plane reference system yet.\n");
    memcpy(refzone, &buf[162], (size_t)6);
    memcpy(projparams, &buf[168], (size_t)360);
    memcpy(coordunits, &buf[528], (size_t)6);
    sscanf(coordunits, "%d", &horz_units);
    if (horz_units != 3) {
	if (horz_units == RADIANS)
	    grunits = radians;
	else if (horz_units == FEET)
	    grunits = feet;
	else if (horz_units == METERS)
	    grunits = meters;
    }
    else
	grunits = arcseconds;
    memcpy(elevunits, &buf[534], (size_t)6);
    sscanf(elevunits, "%d", &vert_units);
    if (vert_units != METERS) {
	if (vert_units == FEET)
	    elunits = feet;
    }
    else
	elunits = meters;
    memcpy(nsides, &buf[540], (size_t)6);
    memcpy(cornercoords, &buf[546], (size_t)192);
    memcpy(elevminmax, &buf[738], (size_t)48);
    memcpy(angle, &buf[786], (size_t)24);
    memcpy(acccode, &buf[810], (size_t)6);
    memcpy(resolutionunits, &buf[816], (size_t)36);
    memcpy(numrowscolumns, &buf[852], (size_t)12);
    memcpy(largestcontour, &buf[864], (size_t)5);
    memcpy(largestunits, &buf[869], (size_t)1);
    memcpy(smallestcontour, &buf[870], (size_t)5);
    memcpy(smallestunits, &buf[875], (size_t)1);
    memcpy(sourcedate, &buf[876], (size_t)4);
    memcpy(revdate, &buf[880], (size_t)4);
    memcpy(insprevflag, &buf[884], (size_t)1);
    memcpy(validationflag, &buf[885], (size_t)1);
    memcpy(suspvoidflags, &buf[886], (size_t)2);
    memcpy(verticaldatum, &buf[888], (size_t)2);
    sscanf(verticaldatum, "%d", &vdatum);
    memcpy(horizontaldatum, &buf[890], (size_t)2);
    sscanf(horizontaldatum, "%d", &hdatum);
    memcpy(dataedition, &buf[892], (size_t)4);
    memcpy(percentvoid, &buf[896], (size_t)4);

    strcpy(fixedcornercoords, cornercoords);
    fixd(fixedcornercoords);
    sscanf(fixedcornercoords, "%lf %lf %lf %lf %lf %lf %lf %lf",
	   &_swe, &_swn, &_nwe, &_nwn, &_nee, &_nen, &_see, &_sen);
    if (horz_units == ARC_SECONDS) {
	_swe /= 3600.0;
	_swn /= 3600.0;
	_nwe /= 3600.0;
	_nwn /= 3600.0;
	_nee /= 3600.0;
	_nen /= 3600.0;
	_see /= 3600.0;
	_sen /= 3600.0;
    }
    strcpy(fixedelevminmax, elevminmax);
    fixd(fixedelevminmax);
    sscanf(fixedelevminmax, "%f %f", &emin, &emax);
    sscanf(resolutionunits, "%12lf%12lf%12lf", &xres, &yres, &zres);

    sscanf(numrowscolumns, "%*d %d", &c); /* number of columns */

    if (refsys == GEOGRAPHIC) { /* this works well with 1 degree DEMs */
        char s1[24], s2[24], s3[24], s4[24], s5[24];
	int d1, d2, d3, d4;
	fscanf(f, "%d %d %d %d", &d1, &d2, &d3, &d4);
	r = d3;
	allocate(r, c);
	fscanf(f, "%24s%24s%24s%24s%24s", s1, s2, s3, s4, s5);
	for (int i = 0; i < c; i++) {
	    if (i > 0) {
	        fscanf(f, "%d %d %d %d", &d1, &d2, &d3, &d4);
		fscanf(f, "%24s%24s%24s%24s%24s", s1, s2, s3, s4, s5);
	    }
	    fixd(s3);
	    float localdatum;
	    sscanf(s3, "%f", &localdatum);
	    for (int j = 0; j < r; j++) {
	        int val;
		fscanf(f, "%d", &val);
		value(j, i, val + int(localdatum));
	    }
	}
	if (p)
	    pclose(f);
	else
	    fclose(f);

	char lon0[80];
	sprintf(lon0, "lon_0=%f", _swe);  // let proj pick the utm zone
	char* parms[4];
	parms[0] = "proj=utm";
	if (hdatum == NULL_HDATUM)
	    parms[1] = "ellps=WGS72";
	parms[2] = lon0;
	parms[3] = "no_defs";
	projection(pj_init(4, parms));
    } /* end of geographic coord DEM */

    else if (refsys == UTM) { /* must be a 7.5 minute DEM */
        /* read entire DEM looking for max number of rows and UTM extent */
        char s1[24], s2[24], s3[24], s4[24], s5[24];
	int d1, d2, d3, d4;
	int val;
	if (_swe < 500000.0) { // west of central meridian - figure 3 users guide
	    float fx1 = _swe / 30.0;
	    if (fx1 - int(fx1) > 0.0)
		fx1 = int(fx1) + 1.0;
	    fx1 *= 30.0;
	    xminutm = int(fx1);// * 30;
	    xmaxutm = xminutm + (c /*- 1*/) * 30;
	    float fymin = _sen / 30.0;
	    if (fymin - int(fymin) > 0.0)
		fymin = int(fymin) + 1.0;
	    yminutm = int(fymin) * 30;
	    float fymax = _nwn / 30.0;
	    ymaxutm = int(fymax) * 30;
	}
	else { // east of central meridian - figure 4 users guide
	    float fx1 = _nwe / 30.0;
	    if (fx1 - int(fx1) > 0.0)
		fx1 = int(fx1) + 1.0;
	    fx1 *= 30;
	    xminutm = int(fx1);
	    xmaxutm = xminutm + (c /*- 1*/) * 30;
	    float fymin = _swn / 30.0;
	    if (fymin - int(fymin) > 0.0)
		fymin = int(fymin) + 1.0;
	    yminutm = int(fymin) * 30;
	    float fymax = _nen / 30.0;
	    ymaxutm = int(fymax) * 30;
	}
	r = ((ymaxutm - yminutm) / 30) + 1;
	allocate(r, c);
	startrow = new int[c];
	endrow = new int[c];
        for (int i = 0; i < c; i++) {
	    fscanf(f, "%d %d %d %d", &d1, &d2, &d3, &d4);
	    fscanf(f, "%24s%24s%24s%24s%24s", s1, s2, s3, s4, s5);
	    fixd(s1); fixd(s2); fixd(s3);
	    float fxgo, fygo;
	    long xgo, ygo;
	    sscanf(s1, "%f", &fxgo);
	    sscanf(s2, "%f", &fygo);
	    xgo = long(fxgo);
	    ygo = long(fygo);
	    float localdatum;
	    sscanf(s3, "%f", &localdatum);
	    int tmprow;
	    for (int j = 0; j < d3; j++) {
	        fscanf(f, "%d", &val);
		long rowutm = ygo + j * 30;
		tmprow = (rowutm - yminutm) / 30;
		if (j == 0)
		    startrow[i] = tmprow;
		if (j == d3 - 1)
		    endrow[i] = tmprow;
		value(tmprow, i, val + int(localdatum));
	    }
	}
	if (p)
	    pclose(f);
	else
	    fclose(f);

	char lon0[80];
	// sprintf(lon0, "lon_0=%f", swe + 1);  // let proj pick the utm zone
	sprintf(lon0, "zone=%d", atoi(refzone));
	char* parms[4];
	parms[0] = "proj=utm";
	//	if (hdatum == NULL_HDATUM)
	//	    parms[1] = "ellps=WGS72";
	switch(hdatum) {
	case UsgsDem::NULL_HDATUM:
	  parms[1] = "ellps=clrk66";
	  break;
	case UsgsDem::NAD27:
	  parms[1] = "ellps=clrk66";
	  break;
	case UsgsDem::WGS72:
	  parms[1] = "ellps=WGS72";
	  break;
	case UsgsDem::WGS84:
	  parms[1] = "ellps=WGS84";
	  break;
	case UsgsDem::NAD83:
	  parms[1] = "ellps=NAD83";
	  break;
	}
	parms[2] = lon0;
	parms[3] = "no_defs";
	projection(pj_init(3/*4*/, parms));
    }
}

void UsgsDem::fixd(char* s) {
    char* d;
    for (d = strchr(s, 'D'); d; d = strchr(s, 'D')) {
	d[0] = 'E';
    }
}

void UsgsDem::print() {
    printf("Name: <%s>\n", name);
    printf("Comment: <%s>\n", comment);
    printf("Process Code: <%s>\n", processcode);
    printf("Section: <%s>\n", section);
    printf("Mapping Center Origin: <%s>\n", mcorigin);
    printf("DEM level: <%s> ", demlevel);
    switch(level) {
    case 1: printf("(DEM-1)\n"); break;
    case 2: printf("(DEM-2)\n"); break;
    case 3: printf("(DEM-3)\n"); break;
    default: printf("(unknown DEM level!)\n"); break;
    }
    printf("Elevation Pattern: <%s>\n", elevpattern);
    printf("Ground Reference System: <%s> ", referencesys);
    switch(refsys) {
    case GEOGRAPHIC:  printf("(Geographic)\n"); break;
    case UTM:         printf("(UTM)\n"); break;
    case STATE_PLANE: printf("(State Plane)\n"); break;
    default:          printf("(unknown reference system!)\n"); break;
    }
    printf("Reference Zone: <%s>\n", refzone);
    printf("Projection Parameters: <%s>\n", projparams);
    printf("Ground Coordinate Units: <%s> ", coordunits);
    switch(horz_units) {
    case RADIANS: printf("(Radians)\n"); break;
    case FEET:    printf("(Feet)\n"); break;
    case METERS:  printf("(Meters)\n"); break;
    case ARC_SECONDS: printf("(ArcSeconds)\n"); break;
    default:          printf("(unknown units!)\n"); break;
    }
    printf("Elevation Units: <%s> ", elevunits);
    switch(vert_units) {
    case RADIANS: printf("(Radians)\n"); break;
    case FEET:    printf("(Feet)\n"); break;
    case METERS:  printf("(Meters)\n"); break;
    case ARC_SECONDS: printf("(ArcSeconds)\n"); break;
    default:          printf("(unknown units!)\n"); break;
    }
    printf("Number of sides: <%s>\n", nsides);
    printf("Corner Coordinates: <%s>\n", cornercoords);
    printf("Corners: SW: %0.1fE %0.1fN; NW: %0.1fE %0.1fN; NE: %0.1fE %0.1fN; SE: %0.1fE %0.1fN\n",
	   	   _swe, _swn, _nwe, _nwn, _nee, _nen, _see, _sen);
    printf("Minimum/Maximum Elevations: <%s>\n", elevminmax);
    printf("Min: %0.2f, Max: %0.2f\n", emin, emax);
    printf("Angle: <%s>\n", angle);
    printf("Accuracy code: <%s>\n", acccode);
    printf("XYZ Resolutions: <%s>\n", resolutionunits);
    printf("XYZ resolutions: %0.1f %0.1f %0.1f\n", xres, yres, zres);
    printf("Rows and Columns: <%s>\n", numrowscolumns);
    printf("Largest Primary Contour Interval: <%s>\n", largestcontour);
    printf("Largest Interval Units: <%s>\n", largestunits);
    printf("Smallest Primary Contour Interval: <%s>\n", smallestcontour);
    printf("Smallest Interval Units: <%s>\n", smallestunits);
    printf("Data source date: <%s>\n", sourcedate);
    printf("Data inspection/revision date: <%s>\n", revdate);
    printf("Inspection/revision flag: <%s>\n", insprevflag);
    printf("Data Validation Flag: <%s>\n", validationflag);
    printf("Suspect and void area flag: <%s>\n", suspvoidflags);
    printf("Vertical datum: <%s>\n", verticaldatum);
    printf("Horizontal datum: <%s>\n", horizontaldatum);
    printf("Data edition: <%s>\n", dataedition);
    printf("Percent void: <%s>\n", percentvoid);
    printf("Number of profiles (columns): %d\n", c);
    printf("Number of rows: %d\n", r);
    if (refsys == GEOGRAPHIC) {
        for (int i = 0; i < cols(); i++) {
	    printf("Column %d:", i);
	    for (int j = 0; j < rows(); j++) {
	        printf(" %d", value(j, i));
	    }
	    printf("\n");
	}
    }
    else if (refsys == UTM) {
        printf("UTM extent: xmin %dE, ymin %dN, xmax %dE, ymax %dN\n",
	       xminutm, yminutm, xmaxutm, ymaxutm);
        for (int i = 0; i < cols(); i++) {
	    printf("Column %d:", i);
	    for (int j = startrow[i]; j <= endrow[i]; j++) {
	        printf(" (row %d)%d", j, value(j, i));
	    }
	    printf("\n");
	}
    }
    else if (refsys == STATE_PLANE)
        printf("can't handle state plane yet.\n");
}

void UsgsDem::writerawpgm(char* path) {
    FILE* f;
    if ((f = fopen(path, "w")) == NULL) {
	printf("couldn't open file for writing\n");
	return;
    }
    fprintf(f, "P5\n%d %d\n", c, r);
    fprintf(f, "%d\n", 255);
    for (int i = r-1; i >= 0; i--) {
	for (int j = 0; j < c; j++) {
	    int v = int(rint(value(i, j) * (255/(emax-emin))));
	    fputc(v, f);
	}
    }
    fclose(f);
}

int UsgsDem::elev_max() {
    return int(rint(emax));
}

int UsgsDem::elev_min() {
    return int(rint(emin));
}

int UsgsDem::value_by_geopos(double x, double y, double& v) {
  // first do a containment test in geo coords
  FPointObj pt(x, y);
  if (grunits == arcseconds) {
    FBoxObj box(_swe, _swn, _nee, _nen);
    if (!box.Contains(pt))
      return 0;
  }
  else if (grunits == meters) {
#if 0
    UV data;
    float x0, y0, x1, y1;
    data.u = min(_swe,_nwe);
    data.v = min(_swn,_sen);
    data = pj_inv(data, _proj);
    x0 = data.u * RAD_TO_DEG;
    y0 = data.v * RAD_TO_DEG;
    data.u = max(_nee,_see);
    data.v = max(_nen,_nwn);
    data = pj_inv(data, _proj);
    x1 = data.u * RAD_TO_DEG;
    y1 = data.v * RAD_TO_DEG;
    FBoxObj box(x0, y0, x1, y1);
    if (!box.Contains(pt))
      return 0;
#endif
  }

  // now find the value by interpolation
  if (grunits == arcseconds) {
    int col = (int(x - _swe) / int(_see - _swe)) * cols();
    int row = (int(y - _swn) / int(_nwn - _swn)) * rows();
    double p0x, p0y, p1x, p1y;
    position(row, col, p0x, p0y);
    position(row+1, col+1, p1x, p1y);
    double param_r = double(y - p0y) / double(p1y - p0y);
    double param_s = double(x - p0x) / double(p1x - p0x);
    double w0 = (1.0 - param_r) * (1.0 - param_s);
    double w1 = param_r * (1.0 - param_s);
    double w2 = (1.0 - param_r) * param_s;
    double w3 = param_r * param_s;
    v = w0 * value(row, col) + w1 * value(row, col+1) +
      w2 * value(row+1, col) + w3 * value(row+1, col+1);
  }
  else if (grunits == meters) {
    UV data;
    float px, py;
    data.u = x * DEG_TO_RAD;
    data.v = y * DEG_TO_RAD;
    data = pj_fwd(data, _proj);
    px = data.u;
    py = data.v;
    long xmin, ymin, xmax, ymax;
    utm_bbox(xmin, ymin, xmax, ymax);
    int col;
    col = int(((px - xmin) / (xmax - xmin)) * cols());
    int row;
    row = int(((py - ymin) / (ymax - ymin)) * rows());
    int p0x = int(xmin + xres * col);
    int p0y = int(ymin + yres * row);
    int p1x = int(xmin + xres * (col + 1));
    int p1y = int(ymin + yres * (row + 1));
    double param_r = double(py - p0y) / double(p1y - p0y);
    double param_s = double(px - p0x) / double(p1x - p0x);
    double w0 = (1.0 - param_r) * (1.0 - param_s);
    double w1 = param_r * (1.0 - param_s);
    double w2 = (1.0 - param_r) * param_s;
    double w3 = param_r * param_s;
    v = w0 * value(row, col) + w1 * value(row, col+1) +
      w2 * value(row+1, col) + w3 * value(row+1, col+1);
  }

  return 1;
}

void UsgsDem::makexyz(int rsize, float sampdist, char* base) {
    double uswe, uswn, usee, usen, unwe, unwn, unee, unen;
    UV data;
    if (reference_system() == GEOGRAPHIC) {
      data.u = _swe * DEG_TO_RAD;
      data.v = _swn * DEG_TO_RAD;
      data = pj_fwd(data, _proj);
      uswe = data.u;
      uswn = data.v;
      data.u = _see * DEG_TO_RAD;
      data.v = _sen * DEG_TO_RAD;
      data = pj_fwd(data, _proj);
      usee = data.u;
      usen = data.v;
      data.u = _nwe * DEG_TO_RAD;
      data.v = _nwn * DEG_TO_RAD;
      data = pj_fwd(data, _proj);
      unwe = data.u;
      unwn = data.v;
      data.u = _nee * DEG_TO_RAD;
      data.v = _nen * DEG_TO_RAD;
      data = pj_fwd(data, _proj);
      unee = data.u;
      unen = data.v;
    }
    else if (reference_system() == UTM) {
      uswe = _swe;
      uswn = _swn;
      usee = _see;
      usen = _sen;
      unwe = _nwe;
      unwn = _nwn;
      unee = _nee;
      unen = _nen;
    }
    else
      return;
    double xmin = Math::min(unwe, uswe);
    double xmax = Math::max(unee, usee);
    double ymin = Math::min(uswn, usen);
    double ymax = Math::max(unwn, unen);
    //    fprintf(stderr, "xmin %0.6f, xmax %0.6f, ymin %0.6f, ymax %0.6f\n",
    //	   xmin, xmax, ymin, ymax);
    int xn = rsize;//512;//1000;//int(xmax - xmin) / 30;
    int yn = rsize;//512;//1000;//int(ymax - ymin) / 30;
    //    fprintf(stderr, "cols %d, rows %d\n", xn, yn);
    //    for (int xi = 0; xi < xn; xi++)
    //      for (int yi = 0; yi < yn; yi++) {

    int xtiles = (Math::round(xmax-xmin) / Math::round(rsize*sampdist))+1;
    int ytiles = (Math::round(ymax-ymin) / Math::round(rsize*sampdist))+1;

    for (int yt = 0; yt < ytiles; yt++) {
      for (int xt = 0; xt < xtiles; xt++) {
	char fname[1024] = "";
	sprintf(fname, "%s.%d.%d.xyz", base ? base : "dted", xt, yt);
	FILE* fp = fopen(fname, "w");
	
	for (int yi = yn; yi >= 0; yi--)
	  for (int xi = 0; xi < xn; xi++) {
	    //	    if (reference_system() == GEOGRAPHIC) {
	    double ux = xmin + (xt*xn*sampdist) + (xi * sampdist);
	    double uy = ymin + (yt*yn*sampdist) + (yi * sampdist);
	    if (xi == 0 && yi == 0)
	      fprintf(stderr, "%s: xmin %0.6f ymin %0.6f\n", fname, ux, uy);
	    data.u = ux;
	    data.v = uy;
	    data = pj_inv(data, _proj);
	    double val = 0.0;
	    value_by_geopos(data.u * RAD_TO_DEG, data.v * RAD_TO_DEG, val);
	    if (isinf(val) || isnan(val))
	      val = 0.0;
	    fprintf(fp, "%0.1f %0.1f %0.6f\n", ux, uy, val);
	    //	printf("%0.6f\n", val);
	    //	printf("%0.1f %0.1f %0.6f\n", xi*30.0, yi*30.0, val);
	    //	    }
#if 0
	    else if (reference_system() == UTM) {
	      int col = xt*xn+xi;
	      int row = yt*yn+yi;
	      double ux, uy;
	      position(row, col, ux, uy);
	      unsigned short val = value(row, col);
	      fprintf(fp, "%0.1f %0.1f %d\n", ux, uy, val);
	    }
#endif
	  }
	fclose(fp);
      }
    }
}

int UsgsDem::geowithin(double x, double y) {
  if (refsys == GEOGRAPHIC) {
  if ((x >= _swe || x >= _nwe) && (x <= _nee || x <= _see) &&
      (y >= _swn || y >= _sen) && (y <= _nen || y <= _nwn))
    return 1;
  else
    return 0;
  }
  else if (refsys == UTM) {
  if ((x >= xminutm) && (x <= xmaxutm) &&
      (y >= yminutm) && (y <= ymaxutm))
    return 1;
  else
    return 0;
  }
}

int UsgsDem::value_by_utmpos(double ux, double uy, double& v) {
  if (refsys == GEOGRAPHIC)
    return DEM::value_by_utmpos(ux, uy, v);
  else if (refsys == UTM) {
    if (geowithin(ux, uy)) {
      int col = int(((ux - xminutm) / (xmaxutm - xminutm)) * cols());
      int row = int(((uy - yminutm) / (ymaxutm - yminutm)) * rows());
      v = value(row, col);
      return 1;
    }
  }
  else
    return 0;
}

void UsgsDem::value_at_utmpos(long ux, long uy, unsigned short v) {
  int col = (int) (float(ux - xminutm) / float(xmaxutm - xminutm)) * cols();
  int row = (int) (float(uy - yminutm) / float(ymaxutm - yminutm)) * rows();
  value(row, col, v);
}
