/*
 * Copyright 1997 Vectaport Inc.
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/* map feature class object */

#ifndef mapfclass_h
#define mapfclass_h

#include <Map/mapfiles.h>
#include <OS/table.h>

class MapGraphic;
class MapCoverage;
class MapFeature;
class MapFeatureClass;
class ostream;

declareTable(MapFeatureTable,int,MapFeature*)

class MapFeatureClass : public MapFiles {
public:
    MapFeatureClass (MapCoverage*, const char* name);
    MapFeatureClass(int nf, MapFeature** flist, const char* name ="");
    MapFeatureClass(int nf);
    virtual ~MapFeatureClass ();

    virtual const char* path() const;
    MapCoverage* coverage() const { return _coverage; }
    virtual void coverage(MapCoverage* cov) { _coverage = cov; }

    int nfeats() const;
    MapFeature* feature(unsigned fnum) const;
    void append(MapFeature*);

    virtual const char* description() const;
    friend ostream& operator << (ostream& s, const MapFeature&);

    boolean initialized() const { return _nfeats>=0; }

    void hidden(void* h) { _hidden = h; }
    void* hidden() { return _hidden; }

    boolean tiled();
    void tiled(boolean);
    virtual void extent(double& x1, double& y1, double& x2, double& y2, int& utmzone);
    void set_extent(double x1, double y1, double x2, double y2, int utmzone) {
      _x1 = x1; _y1 = y1; _x2 = x2; _y2 = y2; _zone = utmzone; }

    // construct junction ans line indices
    void make_junction_idx();
    void make_line_idx();
  
    // access junction and line features by id
    MapFeature* junction_by_id(int id);
    MapFeature* line_by_id(int id);

    // access junction and line features sequentially
    MapFeature* first_junction();
    MapFeature* next_junction();
    MapFeature* first_line();
    MapFeature* next_line();
protected:
    void init_features();

    virtual MapFeature** make_features(int &nfeats) {}
protected:
    MapCoverage* _coverage;

    int _nfeats;
    MapFeature** _features;
  int _size;
    void* _hidden;
    boolean _tiled;

    boolean _extent;
    double _x1;
    double _y1;
    double _x2;
    double _y2;
    int _zone;

  MapFeatureTable* junctionindex;
  MapFeatureTable* lineindex;
  TableIterator(MapFeatureTable)* jit;
  TableIterator(MapFeatureTable)* lit;
};

#endif /* defined(mapfclass_h) */
