/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 * MapFeature -- a geo-located map element
 */

#ifndef mapfeature_h
#define mapfeature_h

#include <OS/enter-scope.h>

class AttributeList;
class Iterator;
class MapDem;
class MapEdge;
class MapFace;
class MapFeatureClass;
class MapNode;
class MapPoint;
class MapPrimitive;
class MapProjection;
class MapRaster;
class MapText;
class ostream;

class MapFeature {
public:
    enum MapFeatureType { PointType, LineType, AreaType, TextType, JunctionType, RasterType, DemType, UnknownType };

    MapFeature(MapFeatureClass* fclass, int id, MapFeatureType);
    MapFeature(int id, MapFeatureType);
    MapFeature();
    MapFeature(MapFeature&);
    virtual ~MapFeature();

    int id() const { return _id; }
    void id(int id) { _id = id; }
    MapFeatureType type() const { return _type; }
    void type(MapFeatureType type) { _type = type; }
    MapFeatureClass* feature_class() const  { return _fclass;}
    void feature_class(MapFeatureClass* fc) { _fclass = fc; }

    AttributeList* attrlist();
    void attrlist(AttributeList* al);
    virtual void create_attrlist() {}
    virtual boolean definition(ostream&) { return true; }
    virtual const char* describe() const;
    virtual float unprojscale();

    virtual char* text_attr(char* attrname);
    virtual float float_attr(char* attrname);
    virtual double double_attr(char* attrname);
    virtual short int short_attr(char* attrname);
    virtual long int long_attr(char* attrname);

    void primitive(MapPrimitive*);
    void edge_primitive(MapEdge*);
    void face_primitive(MapFace*);
    void node_primitive(MapNode*);
    void point_primitive(MapPoint*);
    void text_primitive(MapText*);
    void raster_primitive(MapRaster*);
    void dem_primitive(MapDem*);

    MapPrimitive* primitive() const;
    MapEdge* edge_primitive() const;
    MapFace* face_primitive() const;
    MapNode* node_primitive() const;
    MapPoint* point_primitive() const;
    MapText* text_primitive() const;
    MapRaster* raster_primitive() const;
    MapDem* dem_primitive() const;

    void* graphic() const;
    void graphic(void*);
protected:
    int _id;
    MapFeatureType _type;
    MapFeatureClass* _fclass;
    MapPrimitive* _primitive;
    void* _gr;
    AttributeList* _attrlist;
};

#endif
