/*
 * Copyright 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/*
 * UtmKit - object to construct objects for an UtmEditor
 */

#ifndef utm_kit_h
#define utm_kit_h

#include <MapUnidraw/mapkit.h>

#include <OS/table.h>

#include <InterViews/_enter.h>

class MapFeatureClass;

declareTable(UtmFClassMenuTable,MapFeatureClass*,MenuItem*)
declareTable(UtmFClassTable,MapFeatureClass*,boolean)

class UtmKit : public MapKit {
public:
    UtmKit();
    virtual ~UtmKit();

    virtual Glyph* MakeToolbar();
    virtual Glyph* MakeStates();
    virtual MenuItem* MakeFileMenu();
    virtual MenuItem* MakeRouteMenu();
    virtual MenuItem* MakeMapMenu();
    virtual Glyph* MakeMenus();
    virtual void UpdateLayerMenu();
    virtual void UpdateLibraryMenu();
    void FClassState(MapFeatureClass*, boolean);
    void InitFClassState();
    static UtmKit* Instance();

protected:
    void MakeCheckMenu(MenuItem*, Command*, Command*, const char*,
		       MapFeatureClass*);
    MenuItem* MakeLayerMenus();
    MenuItem* MakeLibraryMenu();
    MenuItem* MakeImportMenu();
    MenuItem* _layermenu;
    MenuItem* _libmenu;
    MenuItem* _mapmenu;
    UtmFClassMenuTable* _fcmenutable;
    UtmFClassTable* _fctable;
    static UtmKit* _utmkit;
};

#include <InterViews/_leave.h>

#endif
