/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/* Class wrapper for vpf feature object */

#ifndef vpf_feature_h
#define vpf_feature_h

#include <Map/mapfeature.h>
#include <Vpf/vpf.h>

#define MISSING_VALUE -2147483647

class VpfFeatureClass;

class VpfFeature : public MapFeature {
public:
    VpfFeature(VpfFeatureClass* fc, int id, MapFeatureType, int,
	       row_type);
    virtual ~VpfFeature();

    VpfFeatureClass* feature_class() const { return (VpfFeatureClass*) _fclass; }

    int ftype() const { return _ftype; }
    row_type row() const { return _row; }
    void* row_element(long int colnumber, void* value, long int* count) const;
    void* named_row_element(char* colname, void* value, long int* count) const;
    virtual boolean definition(ostream&);
    virtual const char* describe() const;
    virtual float unprojscale();

    virtual void create_attrlist();
    virtual char* text_attr(char* attrname);
    virtual float float_attr(char* attrname);
    virtual double double_attr(char* attrname);
    virtual short int short_attr(char* attrname);
    virtual long int long_attr(char* attrname);
    virtual boolean attr_list(ostream&);

    virtual int npts() const { return 0; }
    virtual float * xpoints() const { return nil; }
    virtual float * ypoints() const { return nil; }
    virtual float * zpoints() const { return nil; }
    virtual void origin(float &, float &, float &) const {}
    virtual float xorigin() const { return 0.0; }
    virtual float yorigin() const { return 0.0; }
    virtual float zorigin() const { return 0.0; }
    virtual float xextent() const { return 0.0; }
    virtual float yextent() const { return 0.0; }
    virtual float zextent() const { return 0.0; }

    VpfExtent* extent();
protected:
    int _ftype;
    row_type _row;
    static char _desc[256];
    VpfExtent* _extent;
};

#endif /* defined(vpf_feature_h) */
