/*
 * Copyright 1995 Vectaport Inc., Cider Press
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


#include <VpfUnidraw/vpfcomps.h>
#include <VpfUnidraw/vpfclasses.h>
#include <VpfUnidraw/vpfscripts.h>

#include <Vpf/vpfcov.h>
#include <Vpf/vpfdb.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpflib.h>

#include <MapUnidraw/mapexport.h>

#include <Map/mapproject.h>

#include <OverlayUnidraw/ovclasses.h>
#include <OverlayUnidraw/paramlist.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <Unidraw/editor.h>
#include <Unidraw/iterator.h>
#include <Unidraw/ulist.h>

#include <iostream.h>
#include <stdio.h>
#include <string.h>


/*****************************************************************************/

VpfFClassScript::VpfFClassScript (VpfFClassComp* subj) : MapFClassScript(subj) {}

ClassId VpfFClassScript::GetClassId () { return VPFFCLASS_SCRIPT; }

boolean VpfFClassScript::IsA (ClassId id) {
    return VPFFCLASS_SCRIPT == id || MapFClassScript::IsA(id);
}

int VpfFClassScript::ReadName (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    char name[BUFSIZ];
    ParamList::parse_string(in, name, BUFSIZ);

    if (!in.good()) {
	return -1;
    }
    else {
        
//        *(VpfFeatureClass**)addr1 = new VpfFeatureClass(nil, name);
        ((VpfFClassComp*)addr1)->_tmpname = strdup(name); 
        return 0;
    }
}

boolean VpfFClassScript::Definition (ostream& out) {
  Iterator i;
  boolean status = true;
  Command *cmd = GetCommand();
  ExportScriptType exporttype;

  MapFClassComp* comp = (MapFClassComp*)GetSubject();

  boolean selected = true;
  if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
    exporttype = ((MapExportCmd*)cmd)->type();
    Viewer* viewer = cmd->GetEditor()->GetViewer();
    GraphicView* view = comp->FindView(viewer);
    selected = ((MapExportCmd*)cmd)->Selected(view);
  }
  else {
    MapFeatureClass* fclass = comp->GetFeatureClass();
    out <<  "mapfcl(\"" <<  fclass->name() << "\"" <<
      *comp->GetAttributeList() << ")";
    return status;
  }
    if (selected) {

        Indent(out);
        if (exporttype == MapIdrawType) {
	    MapFeatureClass* fclass = comp->GetFeatureClass();
            const char* name = fclass->name();

            out << "mapfcl(";
	    out << "\"" << name << "\"";

	} else {
            out << "picture(\n";
	}
    }
    Viewer* viewer = cmd->GetEditor()->GetViewer();
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD)) {
	for (First(i); status && !Done(i); ) {
	    OverlayScript* sv = (OverlayScript*)GetView(i);
	    MapFeatureView* view = (MapFeatureView*)comp->FindView(viewer);
	    boolean subselected = ((MapExportCmd*)cmd)->Selected(view);
	    if (subselected) {
		Indent(out);
	        status = sv->Definition(out);
	    }
	    Next(i);
	    if (subselected && !Done(i)) out << ",\n";
	}
    }
    if (selected) {
        FullGS(out);
        Annotation(out);
	Attributes(out);
        out << ")";
    }
    return status;

}    

/*****************************************************************************/

VpfCovScript::VpfCovScript (VpfCovComp* subj) : MapCovScript(subj) {}

ClassId VpfCovScript::GetClassId () { return VPFCOV_SCRIPT; }

boolean VpfCovScript::IsA (ClassId id) {
    return VPFCOV_SCRIPT == id || MapCovScript::IsA(id);
}

int VpfCovScript::ReadName (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    char name[BUFSIZ];
    ParamList::parse_string(in, name, BUFSIZ);

    if (!in.good()) {
	return -1;
    }
    else {
        *(VpfCoverage**)addr1 = new VpfCoverage(nil, name);
        return 0;
    }
}

int VpfCovScript::ReadChildren (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    VpfFClassComp* child = nil;
    VpfCovComp* comp = (VpfCovComp*)addr1;
    char close_paren = ')';
    char key = ':';
    char lookahead;
    char buf[BUFSIZ];

    while (in.good()) {
	ParamList::skip_space(in);
	in.get(lookahead);

	/* keyword section */
	if (lookahead == ':' || lookahead == ')') {
            in.putback(lookahead);
	    break;
	}
	if (lookahead != ',')
            in.putback(lookahead);          
	else
	    ParamList::skip_space(in);
	ParamList::parse_token(in, buf, BUFSIZ);

	if (strcmp(buf,      "mapfcl") == 0)      
	    child = new VpfFClassComp(in);

	else {
	    fprintf(stderr, "unknown fclass object %s, skipping\n", buf);
	    continue;
	}

	if (child != nil) {
	    if (in.good() && child->valid()) {
		comp->Append(child);
	    } else {
	        /* report failure even if one failure */
		delete child;
		return -1;
	    }
	}
    }
    return 0;
}

/*****************************************************************************/

VpfLibScript::VpfLibScript (VpfLibComp* subj) : MapLibScript(subj) {}

ClassId VpfLibScript::GetClassId () { return VPFLIB_SCRIPT; }

boolean VpfLibScript::IsA (ClassId id) {
    return VPFLIB_SCRIPT == id || MapLibScript::IsA(id);
}

int VpfLibScript::ReadName (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    char name[BUFSIZ];
    ParamList::parse_string(in, name, BUFSIZ);

    if (!in.good()) {
	return -1;
    }
    else {
        *(VpfLibrary**)addr1 = new VpfLibrary(nil, name);
        return 0;
    }
}

int VpfLibScript::ReadChildren (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    VpfCovComp* child = nil;
    VpfLibComp* comp = (VpfLibComp*)addr1;
    char close_paren = ')';
    char key = ':';
    char lookahead;
    char buf[BUFSIZ];

    while (in.good()) {
	ParamList::skip_space(in);
	in.get(lookahead);

	/* keyword section */
	if (lookahead == ':' || lookahead == ')') {
            in.putback(lookahead);
	    break;
	}
	if (lookahead != ',')
            in.putback(lookahead);          
	else
	    ParamList::skip_space(in);
	ParamList::parse_token(in, buf, BUFSIZ);

	if (strcmp(buf,      "mapcov") == 0)      
	    child = new VpfCovComp(in);

	else {
	    fprintf(stderr, "unknown coverage object %s, skipping\n", buf);
	    continue;
	}

	if (child != nil) {
	    if (in.good() && child->valid()) {
		comp->Append(child);
	    } else {
	        /* report failure even if one failure */
		delete child;
		return -1;
	    }
	}
    }
    return 0;
}

/*****************************************************************************/

VpfDbScript::VpfDbScript (VpfDbComp* subj) : MapDbScript(subj) {}

ClassId VpfDbScript::GetClassId () { return VPFDB_SCRIPT; }

boolean VpfDbScript::IsA (ClassId id) {
    return VPFDB_SCRIPT == id || MapDbScript::IsA(id);
}

int VpfDbScript::ReadPath (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    char path[BUFSIZ];
    ParamList::parse_string(in, path, BUFSIZ);

    if (!in.good()) {
	return -1;
    }
    else {
        *(VpfDatabase**)addr1 = new VpfDatabase(path, false);
        return 0;
    }
}

int VpfDbScript::ReadChildren (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    VpfLibComp* child = nil;
    VpfDbComp* comp = (VpfDbComp*)addr1;
    char close_paren = ')';
    char key = ':';
    char lookahead;
    char buf[BUFSIZ];

    while (in.good()) {
	ParamList::skip_space(in);
	in.get(lookahead);

	/* keyword section */
	if (lookahead == ':' || lookahead == ')') {
            in.putback(lookahead);
	    break;
	}
	if (lookahead != ',')
            in.putback(lookahead);          
	else
	    ParamList::skip_space(in);
	ParamList::parse_token(in, buf, BUFSIZ);

	if (strcmp(buf,      "maplib") == 0)      
	    child = new VpfLibComp(in);

	else {
	    fprintf(stderr, "unknown library object %s, skipping\n", buf);
	    continue;
	}

	if (child != nil) {
	    if (in.good() && child->valid()) {
		comp->Append(child);
	    } else {
	        /* report failure even if one failure */
		delete child;
		return -1;
	    }
	}
    }
    return 0;
}

/*****************************************************************************/

VpfIdrawScript::VpfIdrawScript (VpfIdrawComp* subj) : MapIdrawScript(subj) {}

ClassId VpfIdrawScript::GetClassId () { return VPF_IDRAW_SCRIPT; }

boolean VpfIdrawScript::IsA (ClassId id) { 
    return VPF_IDRAW_SCRIPT == id || MapIdrawScript::IsA(id);
}

boolean VpfIdrawScript::Emit (ostream& out) {
    Command *cmd = GetCommand();
    MapIdrawComp* comp = (MapIdrawComp*)GetSubject();
    ExportScriptType exporttype;
    if (cmd != nil && cmd->IsA(MAPEXPORT_CMD))
        exporttype = ((MapExportCmd*)cmd)->type();
    else
        exporttype = VpfIdrawType;

    if (exporttype == VpfIdrawType)
        out << "vpfviewer(";
    if (exporttype == MapIdrawType)
        out << "mapdraw(";
    if (exporttype == OvIdrawType)
        out << "drawtool(";

    Iterator i;
    boolean status = true;
    First(i);
    if (!Done(i)) out << "\n";
    for (; status && !Done(i); ) {
	ExternView* ev = GetView(i);
        if (exporttype != OvIdrawType)	
            Indent(out);
        status = ev->Definition(out);
	Next(i);
	if (!Done(i)) out << ",\n";
    }

    const char* proj = comp->GetProjectionName();
    out << "\n" << "    :projection " << "\"" << proj << "\"";

    out << "\n";
    FullGS(out);
    Annotation(out);
    out << ")\n";
    return status;
}

int VpfIdrawScript::ReadChildren (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    VpfDbComp* child = nil;
    VpfIdrawComp* comp = (VpfIdrawComp*)addr1;
    char close_paren = ')';
    char key = ':';
    char lookahead;
    char buf[BUFSIZ];

    while (in.good()) {
	ParamList::skip_space(in);
	in.get(lookahead);

	/* keyword section */
	if (lookahead == ':' || lookahead == ')') {
            in.putback(lookahead);
	    break;
	}
	if (lookahead != ',')
            in.putback(lookahead);          
	else
	    ParamList::skip_space(in);
	ParamList::parse_token(in, buf, BUFSIZ);

	if (strcmp(buf,      "mapdb") == 0)      
	    child = new VpfDbComp(in);

	else {
	    fprintf(stderr, "unknown database object %s, skipping\n", buf);
	    continue;
	}

	if (child != nil) {
	    if (in.good() && child->valid()) {
		comp->Append(child);
	    } else {
	        /* report failure even if one failure */
		delete child;
		return -1;
	    }
	}
    }
    return 0;
}

int VpfIdrawScript::ReadProjection (istream& in, void* addr1, void* addr2, void* addr3, void* addr4) {
    VpfIdrawComp* comp = (VpfIdrawComp*)addr1;
    char proj[BUFSIZ];
    ParamList::parse_string(in, proj, BUFSIZ);

    if (!in.good()) {
	return -1;
    }
    else {
        comp->SetProjectionName(strdup(proj));
        return 0;
    }
}
/*****************************************************************************/

VpfFClassRawScript::VpfFClassRawScript (VpfFClassComp* subj) : MapFClassScript(subj) {}

ClassId VpfFClassRawScript::GetClassId () { return VPFFCLASS_RAW; }

boolean VpfFClassRawScript::IsA (ClassId id) {
    return VPFFCLASS_RAW == id || MapFClassScript::IsA(id);
}

#if 0
static int utmzone(float lon) {
  if (lon > -180.0 && lon < -174.0)
    return 1;
  else if (lon > -174.0 && lon < -168.0)
    return 2;
  else if (lon > -168.0 && lon < -162.0)
    return 3;
  else if (lon > -162.0 && lon < -156.0)
    return 4;
  else if (lon > -156.0 && lon < -150.0)
    return 5;
  else if (lon > -150.0 && lon < -144.0)
    return 6;
  else if (lon > -144.0 && lon < -138.0)
    return 7;
  else if (lon > -138.0 && lon < -132.0)
    return 8;
  else if (lon > -132.0 && lon < -126.0)
    return 9;
  else if (lon > -126.0 && lon < -120.0)
    return 10;
  else if (lon > -120.0 && lon < -114.0)
    return 11;
  else if (lon > -114.0 && lon < -108.0)
    return 12;
  else if (lon > -108.0 && lon < -102.0)
    return 13;
  else if (lon > -102.0 && lon < -96.0)
    return 14;
  else if (lon > -96.0 && lon < -90.0)
    return 15;
  else if (lon > -90.0 && lon < -84.0)
    return 16;
  else if (lon > -84.0 && lon < -78.0)
    return 17;
  else if (lon > -78.0 && lon < -72.0)
    return 18;
  else if (lon > -72.0 && lon < -66.0)
    return 19;
  else if (lon > -66.0 && lon < -60.0)
    return 20;
  else if (lon > -60.0 && lon < -54.0)
    return 21;
  else if (lon > -54.0 && lon < -48.0)
    return 22;
  else if (lon > -48.0 && lon < -42.0)
    return 23;
  else if (lon > -42.0 && lon < -36.0)
    return 24;
  else if (lon > -36.0 && lon < -30.0)
    return 25;
  else if (lon > -30.0 && lon < -24.0)
    return 26;
  else if (lon > -24.0 && lon < -18.0)
    return 27;
  else if (lon > -18.0 && lon < -12.0)
    return 28;
  else if (lon > -12.0 && lon < -6.0)
    return 29;
  else if (lon > -6.0 && lon < 0.0)
    return 30;
  else if (lon > 0.0 && lon < 6.0)
    return 31;
  else if (lon > 6.0 && lon < 12.0)
    return 32;
  else if (lon > 12.0 && lon < 18.0)
    return 33;
  else if (lon > 18.0 && lon < 24.0)
    return 34;
  else if (lon > 24.0 && lon < 30.0)
    return 35;
  else if (lon > 30.0 && lon < 36.0)
    return 36;
  else if (lon > 36.0 && lon < 42.0)
    return 37;
  else if (lon > 42.0 && lon < 48.0)
    return 38;
  else if (lon > 48.0 && lon < 54.0)
    return 39;
  else if (lon > 54.0 && lon < 60.0)
    return 40;
  else if (lon > 60.0 && lon < 66.0)
    return 41;
  else if (lon > 66.0 && lon < 72.0)
    return 42;
  else if (lon > 72.0 && lon < 78.0)
    return 43;
  else if (lon > 78.0 && lon < 84.0)
    return 44;
  else if (lon > 84.0 && lon < 90.0)
    return 45;
  else if (lon > 90.0 && lon < 96.0)
    return 46;
  else if (lon > 96.0 && lon < 102.0)
    return 47;
  else if (lon > 102.0 && lon < 108.0)
    return 48;
  else if (lon > 108.0 && lon < 114.0)
    return 49;
  else if (lon > 114.0 && lon < 120.0)
    return 50;
  else if (lon > 120.0 && lon < 126.0)
    return 51;
  else if (lon > 126.0 && lon < 132.0)
    return 52;
  else if (lon > 132.0 && lon < 138.0)
    return 53;
  else if (lon > 138.0 && lon < 144.0)
    return 54;
  else if (lon > 144.0 && lon < 150.0)
    return 55;
  else if (lon > 150.0 && lon < 156.0)
    return 56;
  else if (lon > 156.0 && lon < 162.0)
    return 57;
  else if (lon > 162.0 && lon < 168.0)
    return 58;
  else if (lon > 168.0 && lon < 174.0)
    return 59;
  else if (lon > 174.0 && lon < 180.0)
    return 60;
  else
    return 0;
}
#endif

boolean VpfFClassRawScript::Definition (ostream& out) {
    Iterator i;
    boolean status = true;
    ExportScriptType exporttype;
    VpfFClassComp* comp = (VpfFClassComp*)GetSubject();
    boolean selected = true;
    exporttype = MapIdrawType;

    MapFeatureClass* fclass = comp->GetFeatureClass();
    const char* name = fclass->name();

    for (First(i); status && !Done(i); ) {
      OverlayScript* sv = (OverlayScript*)GetView(i);
      if (selected && exporttype != OvIdrawType)	    
	Indent(out);
      status = sv->Definition(out);
      Next(i);
      if (selected && !Done(i)) out << ",\n";
    }

    int utm_zone = 0;
    char* parms[4];
    double x1, y1, x2, y2;
    if (!comp->GetAttributeList()->GetAttr("zone")) {
      fclass->extent(x1, y1, x2, y2, utm_zone);
      if (utm_zone == 0) {
	char lon0[80];
	sprintf(lon0, "lon_0=%f", x1);  // let proj pick the utm zone
	parms[0] = "proj=utm";
	parms[1] = "ellps=WGS84";
	parms[2] = lon0;
	parms[3] = "no_defs";

	utm_zone = MapProjection::utmzone(x1+0.1);
      }
    }
    else {
      utm_zone = comp->GetAttributeList()->GetAttr("zone")->Value()->int_val();
      x1 = comp->GetAttributeList()->GetAttr("sw_e")->Value()->double_val();
      y1 = comp->GetAttributeList()->GetAttr("sw_n")->Value()->double_val();
      x2 = comp->GetAttributeList()->GetAttr("ne_e")->Value()->double_val();
      y2 = comp->GetAttributeList()->GetAttr("ne_n")->Value()->double_val();
    }
    out << "\n :proj " << "utm";
    out << " :ellps " << "WGS84";
    out << " :zone " << utm_zone;

    //    if (utm_zone == 0) {
      PJ* proj = pj_init(4, parms);
      UV data;
      data.u = x1 * DEG_TO_RAD;
      data.v = y1 * DEG_TO_RAD;
      data = pj_fwd(data, proj);
      float px1 = data.u;
      float py1 = data.v;
      data.u = x2 * DEG_TO_RAD;
      data.v = y2 * DEG_TO_RAD;
      data = pj_fwd(data, proj);
      float px2 = data.u;
      float py2 = data.v;
      //    }
    out.form("\n :sw_e %.4f :sw_n %.4f :ne_e %.4f :ne_n %.4f",
#if 1
	     px1, py1, px2, py2);  // project extent
#else
	     x1, y1, x2, y2); // already projected extent
#endif

    out << "\n :mapfc \"" << fclass->name() << "\"";
    out << " :mapfc_desc \"" << fclass->description() << "\"";
    MapCoverage* cov = fclass->coverage();
    if (cov) {
        out << " :mapcov \"" << cov->name() << "\"";
        out << " :mapcov_desc \"" << cov->description() << "\"";
        MapLibrary* lib = cov->library();
	if (lib) {
            out << " :maplib \"" << lib->name() << "\"";
            out << " :maplib_desc \"" << lib->description() << "\"";
            MapDatabase* db = lib->database();
	    if (db) {
                out << " :mapdb \"" << db->name() << "\"";
                out << " :mapdb_desc \"" << db->description() << "\"";
  	    }
	}
    }
    return status;
}    

void VpfFClassRawScript::Update() {
    DeleteViews();
    OverlayComp* comps;

    VpfFClassComp* comp = (VpfFClassComp*)GetSubject();
    MapFeatureClass* fclass = comp->GetFeatureClass();
    if (fclass->tiled()) {
      comps = ((VpfFClassComp*)comp)->TileComps()[_tileid];
    }
    else
      comps = GetOverlayComp();

    Iterator i;
    for (comps->First(i); !comps->Done(i); comps->Next(i)) {
        OverlayComp* comp = (OverlayComp*) comps->GetComp(i);
        OverlayScript* sv = CreateRawMapScript(comp);

        if (sv != nil) {
            _views->Append(new UList(sv));
	    SetParent(sv, this);
	}
    }

}

OverlayScript* VpfFClassRawScript::CreateRawMapScript (OverlayComp* comp) {
    OverlayScript* sv = (OverlayScript*) comp->Create(RAWMAP_VIEW);

    if (sv != nil) {
        comp->Attach(sv);
        sv->Update();
    }
    return sv;
}
